"""Core domain data models for LangGate."""

from datetime import UTC, datetime
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, NewType, Self

from pydantic import BaseModel, ConfigDict, Field, SecretStr, model_validator

from langgate.core.fields import NormalizedDecimal

# Type aliases for flexibility while maintaining naming compatibility
ServiceProviderId = NewType("ServiceProviderId", str)
# Model provider might differ from the inference service provider
# The service provider is not intended to be exposed to external consumers of the registry
# The service provider is used by the proxy for routing requests to the correct service
ModelProviderId = NewType("ModelProviderId", str)

# Common model providers for convenience
MODEL_PROVIDER_OPENAI = ModelProviderId("openai")
MODEL_PROVIDER_ANTHROPIC = ModelProviderId("anthropic")
MODEL_PROVIDER_META = ModelProviderId("meta")
MODEL_PROVIDER_GOOGLE = ModelProviderId("google")
MODEL_PROVIDER_MINIMAX = ModelProviderId("minimax")
MODEL_PROVIDER_DEEPSEEK = ModelProviderId("deepseek")
MODEL_PROVIDER_MISTRAL = ModelProviderId("mistralai")
MODEL_PROVIDER_ALIBABA = ModelProviderId("alibaba")
MODEL_PROVIDER_XAI = ModelProviderId("xai")
MODEL_PROVIDER_COHERE = ModelProviderId("cohere")
MODEL_PROVIDER_ELEUTHERIA = ModelProviderId("eleutheria")


class Modality(str, Enum):
    """Supported model modalities."""

    TEXT = "text"
    IMAGE = "image"


class ServiceProvider(BaseModel):
    """Information about a service provider (API service)."""

    id: ServiceProviderId
    base_url: str
    api_key: SecretStr
    default_params: dict[str, Any] = Field(default_factory=dict)


class ModelProviderBase(BaseModel):
    id: ModelProviderId | None = None
    name: str | None = None
    description: str | None = None

    model_config = ConfigDict(extra="allow")


class ModelProvider(ModelProviderBase):
    """Information about a model provider (creator)."""

    id: ModelProviderId = Field(default=...)
    name: str = Field(default=...)


class ContextWindow(BaseModel):
    """Context window information for a model."""

    max_input_tokens: int = 0
    max_output_tokens: int = 0

    model_config = ConfigDict(extra="allow")


class ModelCapabilities(BaseModel):
    """Capabilities of a language model."""

    supports_tools: bool | None = None
    supports_parallel_tool_calls: bool | None = None
    supports_vision: bool | None = None
    supports_audio_input: bool | None = None
    supports_audio_output: bool | None = None
    supports_prompt_caching: bool | None = None
    supports_reasoning: bool | None = None
    supports_response_schema: bool | None = None
    supports_system_messages: bool | None = None

    model_config = ConfigDict(extra="allow")


TokenCost = Annotated[NormalizedDecimal, "TokenCost"]
Percentage = Annotated[NormalizedDecimal, "Percentage"]
TokenUsage = Annotated[NormalizedDecimal, "TokenUsage"]


class BaseModelInfo(BaseModel):
    """Base class for all model types with common fields."""

    id: str | None = None
    name: str | None = None
    provider_id: ModelProviderId | None = None
    description: str | None = None

    model_config = ConfigDict(extra="allow")


class TokenCosts(BaseModel):
    """Token-based cost information for models that charge for input/output tokens."""

    input_cost_per_token: TokenCost = Field(default_factory=Decimal)
    output_cost_per_token: TokenCost = Field(default_factory=Decimal)
    input_cost_per_token_batches: TokenCost | None = None
    output_cost_per_token_batches: TokenCost | None = None
    cache_read_input_token_cost: TokenCost | None = None
    input_cached_cost_per_token: TokenCost | None = None

    model_config = ConfigDict(extra="allow")


class LLMInfoBase(BaseModelInfo):
    costs: TokenCosts | None = None
    capabilities: ModelCapabilities | None = None
    context_window: ContextWindow | None = None


class LLMInfo(LLMInfoBase):
    """Information about a language model."""

    id: str = Field(default=...)  # "gpt-5"
    name: str = Field(default=...)
    provider_id: ModelProviderId = Field(default=...)

    provider: ModelProvider  # Who created it (shown to users)
    description: str | None = None
    costs: TokenCosts = Field(default_factory=TokenCosts)
    capabilities: ModelCapabilities = Field(default_factory=ModelCapabilities)
    context_window: ContextWindow = Field(default_factory=ContextWindow)
    updated_dt: datetime = Field(default_factory=lambda: datetime.now(UTC))

    @model_validator(mode="after")
    def _validate_provider_id(self):
        self.provider_id = self.provider.id
        return self


class ImageGenerationCost(BaseModel):
    """Cost information for image generation with support for multiple pricing models."""

    # For simple flat-rate pricing (most providers)
    flat_rate: Decimal | None = None

    # For dimension-based pricing (OpenAI)
    quality_tiers: dict[str, dict[str, Decimal]] | None = None

    # For usage-based pricing
    cost_per_megapixel: Decimal | None = None
    cost_per_second: Decimal | None = None

    @model_validator(mode="after")
    def validate_exactly_one_pricing_model(self) -> Self:
        """Ensure exactly one pricing model is specified."""
        pricing_models = [
            self.flat_rate,
            self.quality_tiers,
            self.cost_per_megapixel,
            self.cost_per_second,
        ]
        if sum(p is not None for p in pricing_models) != 1:
            raise ValueError("Exactly one pricing model must be set.")
        return self


class ImageModelCost(BaseModel):
    """Cost information for image generation models."""

    token_costs: TokenCosts | None = None
    image_generation: ImageGenerationCost


class ImageModelInfoBase(BaseModelInfo):
    costs: ImageModelCost | None = None


class ImageModelInfo(ImageModelInfoBase):
    """Information about an image generation model."""

    id: str = Field(default=...)
    name: str = Field(default=...)
    provider_id: ModelProviderId = Field(default=...)

    provider: ModelProvider
    description: str | None = None
    costs: ImageModelCost = Field(default=...)
    updated_dt: datetime = Field(default_factory=lambda: datetime.now(UTC))

    @model_validator(mode="after")
    def _validate_provider_id(self) -> Self:
        """Ensure provider_id matches provider.id."""
        self.provider_id = self.provider.id
        return self



"""Shared logging utilities for LangGate."""

import logging
import os
from collections.abc import Callable, Iterator, Mapping, MutableMapping
from contextlib import contextmanager
from logging import StreamHandler
from typing import Any, TextIO, cast

import structlog


class StructLogger(structlog.stdlib.BoundLogger):
    """Custom logger class for structured logging."""


# Default to info if not specified
log_level_str = os.getenv("LOG_LEVEL", "info").lower()
LOG_LEVELS = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}
log_level = LOG_LEVELS.get(log_level_str, logging.INFO)

ProcessorType = Callable[[Any, str, MutableMapping[str, Any]], Mapping[str, Any]]


class MessageIsNormal(logging.Filter):
    def filter(self, record):
        return record.levelno < logging.ERROR


# logs INFO and WARNING to stdout
std_out = logging.StreamHandler()
std_out.setLevel(logging.DEBUG)
std_out.addFilter(MessageIsNormal())

# logs ERROR and CRITICAL to stderr
std_err = logging.StreamHandler()
std_err.setLevel(logging.ERROR)

handlers = [std_out, std_err]

# Modules to suppress DEBUG logs for
info_level_modules = ["httpx"]


def configure_logger(
    json_logs: bool = False, handlers: list[StreamHandler[TextIO]] = handlers
):
    """Configure structlog for the application.

    Args:
        json_logs: Whether to output logs in JSON format (useful for production)
        handlers: List of log handlers to use
    """
    timestamper = structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S")

    logger = structlog.stdlib.get_logger()
    logger.info(
        "logging_configured",
        log_level=log_level_str.upper(),
        json_logs=json_logs,
        py_level=log_level,
    )

    shared_processors: list[ProcessorType] = [
        timestamper,
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.contextvars.merge_contextvars,
        structlog.processors.CallsiteParameterAdder(
            {
                structlog.processors.CallsiteParameter.PATHNAME,
                structlog.processors.CallsiteParameter.FILENAME,
                structlog.processors.CallsiteParameter.MODULE,
                structlog.processors.CallsiteParameter.FUNC_NAME,
                structlog.processors.CallsiteParameter.THREAD,
                structlog.processors.CallsiteParameter.THREAD_NAME,
                structlog.processors.CallsiteParameter.PROCESS,
                structlog.processors.CallsiteParameter.PROCESS_NAME,
            }
        ),
        structlog.stdlib.ExtraAdder(),
    ]

    structlog.configure(
        processors=shared_processors
        + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=StructLogger,
        cache_logger_on_first_use=True,
    )

    logs_render = (
        structlog.processors.JSONRenderer()
        if json_logs
        else structlog.dev.ConsoleRenderer(colors=True)
    )

    _configure_default_logging_by_custom(shared_processors, logs_render, handlers)


def _configure_default_logging_by_custom(
    shared_processors, logs_render, handlers: list[StreamHandler[TextIO]]
):
    """Configure default logging with custom settings."""
    logging.basicConfig(format="%(message)s", level=log_level, handlers=handlers)

    if log_level == logging.DEBUG:
        # suppress DEBUG logs for some modules if log level is DEBUG
        for module in info_level_modules:
            logging.getLogger(module).setLevel(logging.INFO)

    # Use `ProcessorFormatter` to format all `logging` entries.
    formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=shared_processors,
        processors=[
            _extract_from_record,
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            logs_render,
        ],
    )

    for handler in handlers:
        handler.setFormatter(formatter)


def _extract_from_record(_, __, event_dict):
    """Extract thread and process names from record."""
    record = event_dict["_record"]
    event_dict["thread_name"] = record.threadName
    event_dict["process_name"] = record.processName
    return event_dict


@contextmanager
def structlog_contextvars_context(
    context_to_keep: tuple[str, ...] | str = "",
) -> Iterator[dict[str, Any]]:
    """Set fresh context for the duration of a context scope."""
    context = structlog.contextvars.get_contextvars().copy()

    relevant_context = (
        {k: v for k, v in context.items() if k in context_to_keep}
        if context_to_keep
        else {}
    )
    structlog.contextvars.clear_contextvars()
    try:
        yield relevant_context
    finally:
        structlog.contextvars.clear_contextvars()
        structlog.contextvars.bind_contextvars(**context)


@contextmanager
def structlog_add_context(context: dict[str, Any]):
    """Add context to structlog for the duration of a context scope."""
    old_context = structlog.contextvars.get_contextvars().copy()
    try:
        structlog.contextvars.bind_contextvars(**context)
        yield
    finally:
        structlog.contextvars.clear_contextvars()
        structlog.contextvars.bind_contextvars(**old_context)


def is_debug() -> bool:
    """Check if log level is set to DEBUG."""
    return log_level == logging.DEBUG


def get_logger(name: str) -> StructLogger:
    """Get a configured logger instance.

    Args:
        name: The name of the logger, typically __name__

    Returns:
        A structured logger instance
    """
    return cast(StructLogger, structlog.stdlib.get_logger(name))



"""Configuration utilities for LangGate."""

import os
from pathlib import Path
from typing import TypeVar

import yaml
from pydantic import BaseModel, ValidationError

from langgate.core.logging import StructLogger

ConfigSchemaT = TypeVar("ConfigSchemaT", bound=BaseModel)


def resolve_path(
    env_var: str,
    arg_path: Path | None = None,
    default_path: Path | None = None,
    path_desc: str = "path",
    logger: StructLogger | None = None,
) -> Path:
    """Resolve a file path based on priority: args > env > default.

    Args:
        env_var: Environment variable name to check
        arg_path: Path provided in constructor args
        default_path: Default path to use if others not provided
        path_desc: Description for logging
        logger: Optional logger instance for recording path resolution

    Returns:
        Resolved Path object
    """
    # Priority: args > env > default
    resolved_path = arg_path or Path(os.getenv(env_var, str(default_path)))

    # Log the resolved path and its existence
    if logger:
        exists = resolved_path.exists()
        logger.debug(
            f"resolved_{path_desc}",
            path=str(resolved_path),
            exists=exists,
            source="args" if arg_path else ("env" if os.getenv(env_var) else "default"),
        )

    return resolved_path


def load_yaml_config(
    config_path: Path,
    schema_class: type[ConfigSchemaT],
    logger: StructLogger | None = None,
) -> ConfigSchemaT:
    """Load and validate a YAML configuration file using a Pydantic schema.

    Args:
        config_path: Path to the YAML configuration file
        schema_class: The Pydantic schema class to validate against
        logger: Optional logger instance for recording validation results

    Returns:
        The validated schema instance

    Raises:
        FileNotFoundError: If config file doesn't exist
        ValueError: If config file is empty
        yaml.YAMLError: If YAML parsing fails
        ValidationError: If schema validation fails
    """
    try:
        if not config_path.exists():
            if logger:
                logger.error(
                    "config_file_not_found",
                    config_path=str(config_path),
                )
            raise FileNotFoundError(f"Config file not found: {config_path}")

        with open(config_path) as f:
            raw_config = yaml.safe_load(f)

        if raw_config is None:
            if logger:
                logger.error("config_file_is_empty", config_path=str(config_path))
            raise ValueError(f"Config file is empty: {config_path}")

        # Validate configuration using Pydantic schema
        try:
            config = schema_class.model_validate(raw_config)
            if logger:
                logger.info(
                    "loaded_config",
                    config_path=str(config_path),
                )
            return config
        except ValidationError as exc:
            if logger:
                logger.exception(
                    "invalid_config_format",
                    config_path=str(config_path),
                    errors=str(exc),
                )
            raise

    except yaml.YAMLError:
        if logger:
            logger.exception(
                "failed_to_parse_yaml_config", config_path=str(config_path)
            )
        raise
    except Exception:
        if logger:
            logger.exception("failed_to_load_config", config_path=str(config_path))
        raise



"""Schema definitions for YAML configuration validation."""

from typing import Any, Literal

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    SecretStr,
    field_validator,
    model_validator,
)

from langgate.core.fields import UrlOrEnvVar
from langgate.core.models import Modality


class ServiceModelPatternConfig(BaseModel):
    """Configuration for model pattern matching within a service."""

    default_params: dict[str, Any] = Field(default_factory=dict)
    override_params: dict[str, Any] = Field(default_factory=dict)
    remove_params: list[str] = Field(default_factory=list)
    rename_params: dict[str, str] = Field(default_factory=dict)


class ServiceConfig(BaseModel):
    """Configuration for a service provider."""

    api_key: str | SecretStr
    base_url: UrlOrEnvVar | None = None
    api_format: str | None = None
    default_params: dict[str, Any] = Field(default_factory=dict)
    override_params: dict[str, Any] = Field(default_factory=dict)
    remove_params: list[str] = Field(default_factory=list)
    rename_params: dict[str, str] = Field(default_factory=dict)
    model_patterns: dict[str, ServiceModelPatternConfig] = Field(default_factory=dict)

    model_config = ConfigDict(extra="allow")

    @field_validator("api_key")
    def validate_api_key(cls, v):
        """Convert string API keys to SecretStr."""
        if isinstance(v, str):
            # Keep environment variable references as strings
            if v.startswith("${") and v.endswith("}"):
                return v
            return SecretStr(v)
        return v


class ModelServiceConfig(BaseModel):
    """Service configuration for a specific model."""

    provider: str
    model_id: str


class ModelConfig(BaseModel):
    """Configuration for a specific model."""

    id: str
    service: ModelServiceConfig
    modality: Modality | None = None
    name: str | None = None
    description: str | None = None
    api_format: str | None = None
    default_params: dict[str, Any] = Field(default_factory=dict)
    override_params: dict[str, Any] = Field(default_factory=dict)
    remove_params: list[str] = Field(default_factory=list)
    rename_params: dict[str, str] = Field(default_factory=dict)

    model_config = ConfigDict(extra="allow")


class ConfigSchema(BaseModel):
    """Root schema for the configuration YAML."""

    default_params: dict[str, Any] = Field(default_factory=dict)
    services: dict[str, ServiceConfig] = Field(default_factory=dict)
    models: dict[str, list[ModelConfig]] = Field(default_factory=dict)
    app_config: dict[str, Any] = Field(default_factory=dict)
    models_merge_mode: Literal["merge", "replace", "extend"] = "merge"

    @model_validator(mode="after")
    def validate_model_service_providers(self) -> "ConfigSchema":
        """Validate that all model service providers exist in the services configuration."""
        available_providers = set(self.services.keys())

        for modality, model_list in self.models.items():
            for model in model_list:
                if model.service.provider not in available_providers:
                    raise ValueError(
                        f"Model '{model.id}' in modality '{modality}' references service provider '{model.service.provider}' "
                        f"which is not defined in services. Available providers: {sorted(available_providers)}"
                    )

        return self



from decimal import Decimal
from typing import Annotated

from pydantic import (
    AfterValidator,
    AnyHttpUrl,
    PlainSerializer,
    PlainValidator,
    TypeAdapter,
)

AnyHttpUrlAdapter = TypeAdapter(AnyHttpUrl)

# https://github.com/pydantic/pydantic/issues/7186
HttpUrlStr = Annotated[
    str,
    PlainValidator(lambda x: AnyHttpUrlAdapter.validate_strings(x)),
    AfterValidator(lambda x: str(x).rstrip("/")),
]


def validate_url_or_env_var(v: str) -> str:
    """Validate a string as either a URL or an environment variable reference."""
    if v is None:
        return v
    if isinstance(v, str) and v.startswith("${") and v.endswith("}"):
        return v
    return str(AnyHttpUrlAdapter.validate_strings(v)).rstrip("/")


# Type that can be either a URL or an environment variable reference
UrlOrEnvVar = Annotated[str, AfterValidator(validate_url_or_env_var)]


NormalizedDecimal = Annotated[
    Decimal,
    PlainSerializer(lambda v: str(v.normalize()), return_type=str, when_used="json"),
]
